import argparse
import mido
import sys


def version_info():
    try:
        import pkg_resources
        versions = []
        for pkg in ("swood", "pillow-simd", "pillow", "PIL", "pyFFTW", "mido"):
            try:
                versions.append(str(pkg_resources.get_distribution(pkg)))
            except pkg_resources.DistributionNotFound:
                pass
        return " ".join(versions)
    except:
        return "???"


def swoodlive_installed():
    import importlib.util
    return importlib.util.find_spec("swoodlive") is not None


def is_wav(f):
    riff = f.read(4) == b"RIFF"
    f.read(4)
    wave = f.read(4) == b"WAVE"
    f.seek(0)
    return riff and wave


def run_cmd(argv=sys.argv):
    parser = argparse.ArgumentParser(
        description="swood.exe: the automatic ytpmv generator", formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument("infile", type=argparse.FileType("rb"),
                        help="a short wav file to sample as the instrument, or a swood config file")
    parser.add_argument("midi", type=mido.MidiFile,
                        help="the MIDI to play with the wav sample")
    parser.add_argument("output", type=argparse.FileType("wb"),
                        help="path for the output wav file")

    parser.add_argument("--transpose", "-t", type=int,
                        default=0, help="amount to transpose (semitones)")
    parser.add_argument("--speed", "-s", type=float,
                        default=1.0, help="speed multiplier for the MIDI")
    parser.add_argument("--cachesize", "-c", type=float,
                        default=7.5, help="how long to save cached notes")
    parser.add_argument("--binsize", "-b", type=int, default=8192,
                        help="FFT bin size; lower numbers make it faster but more off-pitch")
    parser.add_argument("--fullclip", "-f", action="store_true",
                        help="always use the full sample without cropping")
    parser.add_argument("--no-pbar", "-p", action="store_false",
                        help=argparse.SUPPRESS)

    if swoodlive_installed():
        parser.add_argument("--live",
                            help="listen on a midi input and generate the output in realtime")

    parser.add_argument("--optout", "-o", action="store_true",
                        help="opt out of automatic bug reporting (or set the env variable SWOOD_OPTOUT)")
    parser.add_argument("--version", "-v", action="version", version=version_info(),
                        help="get the versions of swood and its dependencies")

    args = parser.parse_args(argv)

    from . import complain, midiparse, render, sample, soundfont

    with complain.ComplaintFormatter():
        if is_wav(args.infile):
            sample = soundfont.DefaultFont(
                sample.Sample(args.infile, args.binsize))
        else:
            sample = soundfont.SoundFont(args.infile, args)
            # reparse to ensure cli args take precedence over config
            parser.parse_args(argv, args)
        midi = midiparse.MIDIParser(
            args.midi, sample, args.transpose, args.speed)
        renderer = render.NoteRenderer(sample, args.fullclip, args.cachesize)
        renderer.render(midi, args.output, pbar=args.no_pbar)


if __name__ == "__main__":
    run_cmd()



"""Renders music by pitch bending samples according to a MIDI."""

from enum import Enum
import math

from PIL import Image
import progressbar
from numpy import zeros, full, asarray, resize, int32, int64

from . import wavout


class CachedNote:
    """Holds pre-rendered versions of notes, tracking # of uses."""

    def __init__(self, length, rendered, cutoffs):
        self.used = 1
        self.length = length
        self.data = rendered
        self.cutoffs = cutoffs

    def __len__(self):
        return len(self.data)


class FileSaveType(Enum):
    """Enum for selecting where to render to.

    ARRAY_TO_DISK will create a single array and write the contents out when done.
    ARRAY_IN_MEM will render to a single array and return the contents.
    SMART_CACHING will render out to many chunks, flushing them to disk to save memory.
    (If the platform supports it, SMART_CACHING will attempt to memory-map the file instead.)
    It's recommended to use SMART_CACHING (the default from the CLI) over ARRAY_TO_DISK
    as ARRAY_TO_DISK uses lots of memory on any file longer than a few seconds.
    """
    ARRAY_TO_DISK = 0
    ARRAY_IN_MEM = 1
    SMART_CACHING = 2


class NoteRenderer:
    """Renders a WAV file from a MIDI by pitch bending the samples."""

    def __init__(self, sample, fullclip=False, cachesize=7.5, threshold=0.075):
        if threshold < 0:
            return ValueError("The threshold must be a positive number.")
        self.sample = sample
        self.fullclip = fullclip

        self.cachesize = cachesize * sample.framerate
        self.threshold = int(threshold * sample.framerate)

        self.distance_multiplier = (2**32 - 1) / self.threshold * 0.5

        self.notecache = {}

    def zoom(self, img, multiplier):
        """Scales the sound clip (in PIL Image form) by the given multiplier."""
        if multiplier == 1.0:
            return asarray(img, dtype=int32)
        else:
            return asarray(img.resize((int(round(img.size[0] * multiplier)),
                                       self.sample.channels),
                                      resample=Image.BICUBIC), dtype=int32)

    def render_note(self, note):
        """Render a single note and return an array (with optional cutoffs)."""
        instrument = note.instrument

        if instrument.sample is None:
            return None, None

        if instrument.noscale:
            scaled = self.zoom(instrument.sample.img, 1.0)
        else:
            scaled = self.zoom(instrument.sample.img,
                               instrument.sample.fundamental_freq / note.pitch)
        if self.fullclip or instrument.fullclip:
            return scaled, full(instrument.sample.channels, scaled.shape[1], dtype=int32)

        # cache variables for faster lookups
        # see https://stackoverflow.com/q/37202463
        length = note.length
        channels = instrument.sample.channels

        # get the area on the end of the clip that it's ok to cut off at
        if scaled.shape[1] > length:
            distance_multiplier = self.distance_multiplier
            note_ending = scaled[:, length:length + self.threshold]
        else:
            distance_multiplier = -self.distance_multiplier
            start = min(0, length - self.threshold)
            note_ending = scaled[start:]

        # find the closest zero crossing within the threshold & cut off there
        # removes "clicking" sounds from the audio suddenly cutting out
        cutoffs = zeros(channels, dtype=int32)
        cutoff_scores = full(channels, 2**63 - 1, dtype=int64)
        for channel, audio in enumerate(note_ending):
            for distance, val in enumerate(audio):
                score = abs(val) + (distance * distance_multiplier)
                if score < cutoff_scores[channel]:
                    cutoff_scores[channel] = score
                    cutoffs[channel] = distance
        cutoffs += length
        return scaled, cutoffs

    def render(self, midi, filename=None, pbar=False, savetype=FileSaveType.SMART_CACHING, clear_cache=True):
        """Renders from a MIDIParser to an array or WAV file using Samples.

        Args:
            midi: The (pre-parsed) MIDI file to render.
            filename: A file or file path to save the WAV file to. Not needed with
            FileSaveType.ARRAY_IN_MEM.
            pbar: Show a progress bar on STDOUT while rendering. Defaults to False.
            savetype: The FileSaveType to use when writing the sound data. Defaults
            to FileSaveType.SMART_CACHING.
            clear_cache: Remove all notes from the temporary cache after rendering
            the MIDI. It's recommended to disable this if you're rendering many MIDIs
            and have memory to spare. Defaults to True.
        """

        if savetype != FileSaveType.ARRAY_IN_MEM and filename is None:
            return ValueError("When not outputting to an array in memory, you need to specify a filename.")
        if self.fullclip:
            # leave a small buffer at the end with space for one more sample
            output_length = midi.length + \
                int(math.ceil(midi.maxpitch * len(self.sample)))
        else:
            # it has to cut off sounds at the threshold anyway
            output_length = midi.length + self.threshold

        if savetype == FileSaveType.SMART_CACHING:
            output = wavout.CachedWavFile(
                output_length, filename, self.sample.framerate, self.sample.channels)
        else:
            output = wavout.UncachedWavFile(
                output_length, filename, self.sample.framerate, self.sample.channels)

        if pbar:
            bar = progressbar.ProgressBar(widgets=[progressbar.Percentage(
            ), " ", progressbar.Bar(), " ", progressbar.ETA()], max_value=midi.notecount)
            update = bar.update
            progress = 0

        # "inlining" these variables can speed up the lookup, making it faster
        # see https://stackoverflow.com/questions/37202463
        caching = self.cachesize > 0
        add_data = output.add_data
        maxvolume = midi.maxvolume
        cachesize = self.cachesize

        if caching:
            tick = 8
            notecache = self.notecache

        for time, notes in midi.notes:
            for note in notes:
                if note in notecache:
                    rendered_note = notecache[note]
                    rendered_note.used += 1  # increment the used counter each time for the "GC" below
                else:
                    rendered_note = CachedNote(time, *self.render_note(note))
                    notecache[note] = rendered_note
                if rendered_note.data is not None and rendered_note.data.shape[0] != 0:
                    if note.instrument.pan == 0.5:
                        add_data(time, rendered_note.data *
                                 (note.volume / midi.maxvolume *
                                  note.instrument.volume),
                                 rendered_note.cutoffs)
                    else:
                        add_data(time, rendered_note.data *
                                 (note.volume / midi.maxvolume *
                                  note.instrument.volume),
                                 rendered_note.cutoffs,
                                 ((1 - note.instrument.pan) * 2, note.instrument.pan * 2))
                if pbar:
                    # increment progress bar
                    progress += 1
                    update(progress)

            if caching:
                # cache "garbage collection":
                # if a CachedNote is more than <cachesize> seconds old and not
                # used >2 times it removes it from the cache to save mem(e)ory
                tick += 1
                if tick == 15:
                    tick = 0
                    for k in list(notecache.keys()):
                        if time - notecache[k].length > cachesize and notecache[k].used < 3:
                            del notecache[k]

        if caching and clear_cache:
            notecache.clear()

        if savetype == FileSaveType.ARRAY_IN_MEM:
            return output.channels
        else:
            output.save()



"""Parses .swood files."""

# See https://github.com/milkey-mouse/swood/issues/1

from os.path import normpath, abspath, dirname, join
from collections import defaultdict
from enum import Enum

from PIL.Image import BILINEAR
from .sample import Sample
from .instruments import *
from . import complain
import zipfile
import string

# user-friendly repr for zip-loaded samples
zipfile.ZipExtFile.__repr__ = (
    lambda self: "<zipped WAV file '{}'>".format(self.name))


class SoundFontSyntaxError(complain.ComplainToUser, SyntaxError):
    """Tells the user when something is wrong with the config file."""

    def __init__(self, line, line_text, error_desc):
        self.line = line
        self.line_text = line_text
        self.error_desc = error_desc

    def __str__(self):
        return "Syntax error on line {}:\n".format(self.line + 1) + \
               self.line_text + "\n" + self.error_desc


class Instrument:
    """Holds information about a MIDI instrument or track."""

    def __init__(self, fullclip=False, noscale=False, sample=None, volume=0.9, pan=0.5):
        self.fullclip = fullclip
        self.noscale = noscale
        self.sample = sample
        self.volume = volume
        self.pan = pan

    def __hash__(self):
        if isinstance(self.sample, Sample):
            return hash((self.noscale, self.sample.filename, self.volume, self.pan))
        else:
            return hash((self.noscale, None, self.volume, self.pan))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __repr__(self):
        return "Instrument(noscale={}, sample={}, volume={}, pan={})".format(self.noscale, self.sample, self.volume, self.pan)


class SoundFont:
    """Parses and holds information about .swood files."""

    def __init__(self, filename, arguments):
        self.arguments = arguments
        self.load_instruments()
        self.samples = set()
        self.channels = {}

        if isinstance(filename, str):
            self.file = open(filename)
        elif filename is not None:
            self.file = filename

        if filename is not None:
            if zipfile.is_zipfile(self.file):
                self.file = zipfile.ZipFile(self.file)
                self.load_zip()
                self.load_samples_from_zip()
            else:
                self.load_ini()
                self.load_samples_from_txt()

    def load_instruments(self):
        self.instruments = defaultdict(list)
        self.percussion = defaultdict(list)
        for names in instruments:
            new_instrument = Instrument()
            for name in names:
                if isinstance(name, str):
                    name = name.lower()
                self.instruments[name].append(new_instrument)
                self.instruments["non-percussion"].append(new_instrument)
            self.instruments["all"].append(new_instrument)
        # percussion is a bit weird as it doesn't actually use MIDI instruments;
        # any event on channel 10 is percussion, and the actual instrument is
        # denoted by the note number (with valid #s ranging 35-81).
        for idx, *names in percussion:
            new_instrument = Instrument(fullclip=True, noscale=True)
            self.percussion[idx].append(new_instrument)
            for name in names:
                if isinstance(name, str):
                    name = name.lower()
                self.percussion[name].append(new_instrument)
                self.percussion["percussion"].append(new_instrument)
            self.instruments["all"].append(new_instrument)

    def load_ini(self):
        self.file.seek(0)
        if "b" in self.file.mode:
            self.parse(self.file.read().decode("utf-8"))
        else:
            self.parse(self.file.read())

    def load_zip(self):
        """Parses a ZIP of a .swood INI file and its samples without extracting."""
        try:
            valid_extensions = {"swood", "ini", "txt"}
            ini_path = next(fn for fn in self.file.namelist()
                            if fn.split(".")[-1] in valid_extensions)
        except StopIteration:
            raise complain.ComplainToUser(
                "Couldn't find config file in ZIP. Be sure it ends in .ini, .swood, or .txt.'")
        config_txt = self.file.read(ini_path)
        tokenize(config_txt)

    def strip_comments(self, line):
        hash_index = line.find("#")
        if hash_index == -1:
            return line.strip(string.whitespace + "\n")
        else:
            return line[:hash_index].strip(string.whitespace + "\n")

    def parse(self, config):
        affected_instruments = []
        parse_arguments = None
        for linenum, raw_text in enumerate(config.replace("\r\n", "\n").split("\n")):
            text = self.strip_comments(raw_text)
            if text == "":
                continue
            elif text.startswith("[") and text.endswith("]"):
                header_name = text[1:-1].lower()
                if header_name in ("arguments", "args", "options"):
                    affected_instruments = []
                    parse_arguments = True
                elif header_name in ("default", "all"):
                    affected_instruments = self.instruments["all"]
                    parse_arguments = False
                elif header_name in self.instruments:
                    affected_instruments = self.instruments[header_name]
                    parse_arguments = False
                elif header_name in self.percussion:
                    affected_instruments = self.percussion[header_name]
                    parse_arguments = False
                elif header_name in ("non percussion", "nonpercussion"):
                    affected_instruments = self.percussion["non-percussion"]
                    parse_arguments = False
                elif len(header_name) == 3 and header_name.startswith("p"):
                    try:
                        affected_instruments = \
                            self.percussion[int(header_name[1:])]
                        parse_arguments = False
                    except (ValueError, KeyError):
                        raise SoundFontSyntaxError(
                            linenum, raw_text, "Header not recognized.")
                else:
                    raise SoundFontSyntaxError(
                        linenum, raw_text, "Header not recognized.")
            elif "=" in text:
                parts = text.split("=")
                name = parts[0].strip()
                value = parts[1]
                if parse_arguments is None:
                    raise SoundFontSyntaxError(
                        linenum, raw_text,
                        "No header specified. For defaults, specify '[default]' on the line before."
                    )
                elif parse_arguments:
                    possible_args = {
                        "transpose": int,
                        "speed": float,
                        "cachesize": float,
                        "binsize": int,
                    }
                    if name in possible_args:
                        try:
                            setattr(self.arguments, name,
                                    possible_args[name](value))
                        except ValueError:
                            raise SoundFontSyntaxError(
                                linenum, raw_text, "'{}' is not a valid value for '{}'".format(value, name))
                elif name in ("file", "sample"):
                    for instrument in affected_instruments:
                        if value.lower() in ("", "none", "null"):
                            instrument.sample = None
                        else:
                            instrument.sample = value
                            self.samples.add(value)
                elif name in ("volume", "vol"):
                    for instrument in affected_instruments:
                        try:
                            instrument.volume = int(value) / 100
                            if instrument.volume > 0.95:
                                print(
                                    "Warning: Volumes higher than 95 may cause clipping or other glitches")
                        except ValueError:
                            raise SoundFontSyntaxError(
                                linenum, raw_text, "'{}' is not a valid number".format(value))
                elif name == "pan":
                    for instrument in affected_instruments:
                        try:
                            pan = float(value)
                        except ValueError:
                            raise SoundFontSyntaxError(
                                linenum, raw_text, "'{}' is not a valid number".format(value))
                            if pan < 0 or pan > 1:
                                raise SoundFontSyntaxError(
                                    linenum, raw_text, "'{}' is outside of the allowed 0.0-1.0 range".format(value))
                            else:
                                instrument.pan = pan
                        instrument.pan = float(value)
                elif name == "fullclip":
                    for instrument in affected_instruments:
                        if value.lower() == "true":
                            instrument.fullclip = True
                        elif value.lower() == "false":
                            instrument.fullclip = False
                        else:
                            raise SoundFontSyntaxError(linenum, raw_text,
                                                       "fullclip must be 'True' or 'False'; '{}' is invalid".format(value))
                elif name == "noscale":
                    for instrument in affected_instruments:
                        if value.lower() == "true":
                            instrument.noscale = True
                        elif value.lower() == "false":
                            instrument.noscale = False
                        else:
                            raise SoundFontSyntaxError(linenum, raw_text,
                                                       "noscale must be 'True' or 'False'; '{}' is invalid".format(value))
                else:
                    raise SoundFontSyntaxError(
                        linenum, raw_text, "'{}' is not a valid property".format(name))

    def wavpath(self, relpath):
        # only works on non-zip files
        return normpath(join(dirname(abspath(self.file.name)), relpath))

    def load_samples_from_txt(self):
        loaded_samples = {}
        for fn in self.samples:
            loaded_samples[fn] = Sample(
                self.wavpath(fn),
                self.arguments.binsize
            )
        for instruments in self.instruments.values():
            for instrument in instruments:
                if isinstance(instrument.sample, str):
                    instrument.sample = loaded_samples[instrument.sample]
        self.add_samples(loaded_samples)

    def load_samples_from_zip(self):
        loaded_samples = {}
        for fn in self.samples:
            filepath = normpath(join(dirname(abspath(self.file.fp.name)), fn))
            with self.file.open(filepath) as zipped_wav:
                loaded_samples[fn] = Sample(zipped_wav, self.arguments.binsize)
        self.add_samples(loaded_samples)

    def add_samples(self, loaded_samples):
        self.framerate = max(s.framerate for s in loaded_samples.values())
        self.channels = max(s.channels for s in loaded_samples.values())
        self.length = max(len(s) for s in loaded_samples.values())
        for samp in loaded_samples.values():
            multiplier = self.framerate / samp.framerate
            samp._img = samp.img.resize(
                (int(round(samp.img.size[0] * multiplier)), samp.channels),
                resample=BILINEAR)
        for instruments in self.instruments.values():
            for instrument in instruments:
                if isinstance(instrument.sample, str):
                    instrument.sample = loaded_samples[instrument.sample]
        if self.channels != 2:
            warned_pan = False
            for instruments in self.instruments.values():
                for instrument in instruments:
                    if instrument.pan != 0.5:
                        instrument.pan = 0.5
                        if not warned_pan:
                            print("Warning: Audio has >2 channels; pan ignored")
                            warned_pan = True

    def __len__(self):
        return self.length


def DefaultFont(samp):
    sf = SoundFont(None, None)
    sf.framerate = samp.framerate
    sf.channels = samp.channels
    sf.length = samp.length
    for instruments in sf.instruments.values():
        for instrument in instruments:
            instrument.sample = samp
    return sf



"""User-friendly exception handler for swood."""

import traceback
import sys
import os


class ComplainToUser(Exception):
    """When used with ComplaintFormatter, tells the user what error (of theirs) caused the failure and exits."""
    pass


class ComplaintFormatter(object):
    """Notifies the user when the program fails predictably and uploads bug reports.

    When used in a with statement, ComplaintFormatter catches all exceptions. If the
    exception is a ComplainToUser exception, it will simply print the error message
    and exit (with an exit code of 1). If the exception is something else (i.e. an
    actual, unexpected exception), it will upload the traceback to the swood debug
    server (unless the user has opted out of sending bug reports.)
    """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc, tb):
        if isinstance(exc, ComplainToUser):
            print("Error: {}".format(exc))
            sys.exit(1)
        elif isinstance(exc, Exception):
            tb = traceback.format_exc()
            if "--optout" in sys.argv or "-o" in sys.argv or os.environ.get("SWOOD_OPTOUT") is not None:
                print(
                    "Something went wrong. A bug report will not be sent because of your environment variable/CLI option.")
                traceback.print_exc()
            else:
                print(
                    "Something went wrong. A bug report will be sent to help figure it out. (see --optout)")
                try:
                    import http.client
                    conn = http.client.HTTPSConnection("meme.institute")
                    conn.request("POST", "/swood/bugs/submit", tb)
                    resp = conn.getresponse().read().decode("utf-8")
                    if resp == "done":
                        print("New bug submitted!")
                    elif resp == "dupe":
                        print("This bug is already in the queue to be fixed.")
                    else:
                        raise Exception
                except Exception:
                    print(
                        "Apparently we can't even send a bug report right. Sorry about that.")
                    traceback.print_exc()
        return True



from numpy import zeros, int32, fromfile, full, ndarray
from numpy import dtype as dtype_info
from collections import defaultdict
from . import complain
import mmap
import wave
import os


class UncachedWavFile:
    """Creates a single large array for the output and writes it to disk at the end."""

    def __init__(self, length, filename, framerate, channels=1, dtype=int32):
        self.channels = zeros((channels, length), dtype=dtype)
        self.framerate = framerate
        self.filename = filename

    def add_data(self, start, data, cutoffs=None, volumes=None):
        """Add sound data at a specified position.

        Args:
            start: How many samples into the output the data should start.
            data: A NumPy array of data to add to the output.
            cutoffs: An array of integers that specifies where to cut off each channel. (optional)
            volumes: An array of floats to multiply each channel's data by. (optional)
        """
        if cutoffs is None:
            cutoffs = full(self.channels.shape[0],
                           data.shape[1], dtype=int32)
        for chan in range(self.channels.shape[0]):
            selectChan = min(chan, data.shape[0])
            length = min(self.channels.shape[1] - start,
                         cutoffs[selectChan],
                         data.shape[1])
            if volumes is None:
                self.channels[chan][start:start + length] += \
                    data[selectChan][:length].astype(self.channels.dtype)
            else:
                self.channels[chan][start:start + length] += \
                    data[selectChan][:length].astype(self.channels.dtype) * \
                    volumes[selectChan]

    def save(self):
        """Write the output array to the file."""
        try:
            # open the file manually to catch I/O exceptions
            # as IOError instead of wave.Error
            with (open(self.filename, "wb") if isinstance(self.filename, str) else self.filename) as wavfile:
                with wave.open(wavfile, "w") as wav:
                    wav.setparams((self.channels.shape[0],  # channels
                                   self.channels.dtype.itemsize,  # sample width
                                   self.framerate,  # sample rate
                                   self.channels.shape[1],  # number of frames
                                   "NONE", "not compressed"))  # compression type (none are supported)
                    wav.writeframesraw(self.channels.flatten(order="F"))
        except IOError:
            raise complain.ComplainToUser(
                "Can't save output file '{}'.".format(self.filename))


def CachedWavFile(*args, **kwargs):
    """Tries to create a MemMapWavFile and falls back to ChunkedWavFile."""

    # Windows doesn't like mmap'ing as non-admin
    if os.name == "nt":
        try:
            import ctypes
            if ctypes.windll.shell32.IsUserAnAdmin() == 0:
                return ChunkedWavFile(*args, **kwargs)
        except:
            pass

    try:
        return MemMapWavFile(*args, **kwargs)
    except PermissionError:
        return ChunkedWavFile(*args, **kwargs)
    except Exception as e:
        # Probably has more obscure errors here so just ignore them
        return ChunkedWavFile(*args, **kwargs)  # shh bby is ok


class MemMapWavFile:
    """Uses memory mapped arrays to easily cache WAV files."""

    def __init__(self, length, filename, framerate, channels=1, dtype=int32):
        # this was so much simpler to write than ChunkedWavFile
        # really wish i learned about memmap before writing it
        if isinstance(filename, str):
            self.wavfile = open(filename, "wb+")
            self._auto_close = True
        else:
            self.wavfile = filename
            self._auto_close = False

        wav = wave.Wave_write(self)
        wav.close = lambda: None
        wav.initfp(self.wavfile)
        wav.setparams((channels, dtype_info(dtype).itemsize,
                       framerate, length, "NONE", "not compressed"))
        wav._write_header(length)
        del wav
        self.wavfile.flush()
        self.wav_memmap = mmap.mmap(self.wavfile.fileno(),
                                    length * dtype_info(dtype).itemsize,
                                    access=mmap.ACCESS_WRITE)
        self.channels = ndarray((channels, length), dtype,
                                self.wav_memmap, self.wavfile.tell(), order="F")
        self.framerate = framerate
        self.filename = filename

    def add_data(self, start, data, cutoffs=None, volumes=None):
        """Add sound data at a specified position.

        Args:
            start: How many samples into the output the data should start.
            data: A NumPy array of data to add to the output.
            cutoffs: An array of integers that specifies where to cut off each channel. (optional)
            volumes: An array of floats to multiply each channel's data by. (optional)
        """
        if cutoffs is None:
            cutoffs = full(self.channels.shape[0],
                           data.shape[1], dtype=int32)
        for chan in range(self.channels.shape[0]):
            selectChan = min(chan, data.shape[0])
            length = min(self.channels.shape[1] - start,
                         cutoffs[selectChan],
                         data.shape[1])
            if volumes is None:
                self.channels[chan][start:start + length] += \
                    data[selectChan][:length].astype(self.channels.dtype)
            else:
                self.channels[chan][start:start + length] += \
                    data[selectChan][:length].astype(self.channels.dtype) * \
                    volumes[selectChan]

    def save(self):
        del self.wav_mmap  # the way to close a memmap is to delete it
        if self._auto_close:
            self.wavfile.close()


class defaultdictkey(defaultdict):
    """Variation of collections.defaultdict that passes the key into the factory."""

    def __missing__(self, key):
        self[key] = self.default_factory(key)
        return self[key]


class ChunkedWavFile:
    """Uses chunks of data to efficiently write a WAV file to disk without storing a large array."""

    def __init__(self, length, filename, framerate, channels=1, chunksize=32768, dtype=int32):
        # 32768 chunk size holds ~1/6 second at 192khz
        # and ~0.75 seconds at 44.1khz (cd quality)
        self.framerate = framerate
        self.channels = channels
        self.chunksize = chunksize

        self.dtype = dtype
        self.itemsize = dtype_info(self.dtype).itemsize
        self.chunkspacing = self.channels * self.chunksize * self.itemsize

        self.saved_to_disk = set()
        self.chunks = defaultdictkey(self._create_chunk)

        if isinstance(filename, str):
            self.wavfile = open(filename, "wb+")
            self._auto_close = True
        else:
            self.wavfile = filename
            self._auto_close = False

        self.wav = wave.Wave_write(self)
        self.wav.close = lambda: None
        self.wav.initfp(self.wavfile)

        self.wav.setparams((channels, self.itemsize, self.framerate,
                            0, "NONE", "not compressed"))
        self.wav._write_header(0)  # start at 0 length and patch header later
        self._header_length = self.wavfile.tell()

    def _create_chunk(self, key):
        """Create a new chunk filled with zeros, or load one from disk."""
        if key in self.saved_to_disk:
            return self._load_chunk(key)
        else:
            return zeros((self.channels, self.chunksize), dtype=self.dtype)

    def _load_chunk(self, idx):
        """Load a chunk from disk into the cache."""
        raise NotImplementedError  # not needed for linear writes and it's broken
        self.saved_to_disk.discard(idx)
        self.wavfile.seek(self._header_length + (self.chunkspacing * idx))
        raw_data = fromfile(self.wavfile, dtype=self.dtype,
                            count=self.channels * self.chunksize)
        return raw_data.reshape((self.channels, self.chunksize), order="F")

    def _save_chunk(self, idx):
        """Write a chunk to disk and remove it from the cache."""
        self.wavfile.seek(self._header_length + (self.chunkspacing * idx))
        self.chunks[idx].flatten(order="F").tofile(self.wavfile)
        self.saved_to_disk.add(idx)
        del self.chunks[idx]

    def flush_cache(self, to_idx=None):
        """Save all (or all up to a certain index) chunks in memory to disk and remove them fron the cache."""
        # we should still sort the keys even though it's theoretically not needed
        # because sequential disk writes are faster on both SSDs and hard disks
        for idx in sorted(self.chunks.keys()):
            if to_idx is not None and idx >= to_idx:
                break
            else:
                self._save_chunk(idx)

    def fill_empty_chunks(self):
        """Fill all the unwritten chunks with zeros.

        This isn't strictly necessary, as on POSIX and Windows seek()ing past
        the end of a file fills the unwritten part with zeros. Unfortunately,
        doing so is still undefined behavior and should be avoided. This function
        iterates over the unwritten chunks, filling them with zeros."""
        try:
            for chunk_idx in range(max(self.saved_to_disk)):
                if chunk_idx not in self.saved_to_disk:
                    # the chunk won't exist but it will create it full of zeros
                    self._save_chunk(chunk_idx)
        except ValueError:  # if no chunks are stored to disk max() will fail
            pass

    def add_data(self, start, data, cutoffs=None, volumes=None):
        """Add sound data at a specified position.

        Args:
            start: How many samples into the output the data should start.
            data: A NumPy array of data to add to the output.
            cutoffs: An array of integers that specifies where to cut off each channel. (optional)
            volumes: An array of floats to multiply each channel's data by. (optional)
        """
        if cutoffs is None:
            cutoffs = full(self.channels, data.shape[1], dtype=int32)

        if volumes is not None:
            for chan in range(data.shape[0]):
                data[chan] *= volumes[chan]

        data = data.astype(self.dtype)

        chunksize = self.chunksize
        chunk_start = start // chunksize
        chunk_offset = start - (chunk_start * chunksize)
        for chan in range(self.channels):
            selectChan = min(chan, data.shape[0])
            current_chunk = chunk_start
            cutoff = min(cutoffs[selectChan], len(data[selectChan]))
            if cutoff + chunk_offset <= chunksize:
                self.chunks[current_chunk][chan][chunk_offset:chunk_offset + cutoff] += \
                    data[selectChan][:cutoff]
            else:
                self.chunks[current_chunk][chan][chunk_offset:] += \
                    data[selectChan][:chunksize -
                                     chunk_offset]
                bytes_remaining = cutoff - chunksize + chunk_offset
                current_chunk += 1
                while bytes_remaining >= chunksize:
                    self.chunks[current_chunk][chan] += \
                        data[selectChan][
                            cutoff - bytes_remaining:cutoff - bytes_remaining + chunksize]
                    current_chunk += 1
                    bytes_remaining -= chunksize
                self.chunks[current_chunk][chan][:bytes_remaining] += \
                    data[selectChan][cutoff - bytes_remaining:cutoff]
        self.flush_cache(chunk_start)

    def save(self):
        """Flush the cache of chunks to disk, patch the WAV header with the new length, and close the file."""
        self.flush_cache()
        self.fill_empty_chunks()
        try:
            self.wav._datawritten = (
                max(self.saved_to_disk) + 1) * self.chunkspacing
        except ValueError:
            self.wav._datawritten = 0
        self.wav._patchheader()
        if self._auto_close:
            self.wavfile.close()



"""Parses MIDI files into a list of notes in chronological order."""

from copy import copy
import collections
import operator

import mido

from . import complain, soundfont
from .sample import Sample


def note_to_freq(notenum):
    """Converts a MIDI note number to a frequency.

    See https://en.wikipedia.org/wiki/MIDI_Tuning_Standard
    """
    return (2.0 ** ((notenum - 69) / 12.0)) * 440.0


class Note:
    """Holds information about each MIDI note."""

    def __init__(self, volume=127, start=0, pitch=0, instrument=1, percussion=False):
        self.instrument = instrument
        self.percussion = percussion

        self.volume = volume
        self.start = start
        self.pitch = pitch
        self.length = 0

    def __hash__(self):
        return hash((self.length, self.pitch, self.start, hash(self.instrument), self.percussion))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return "Note(length={}, volume={}, start={}, pitch={})".format(self.length, self.volume, self.start, self.pitch)

    def finalize(self, time):
        if not self.percussion:
            self.pitch = note_to_freq(self.pitch)
        self.length = time - self.start
        return self


class MIDIParser:
    """Parses a MIDI file into a chronological list of notes."""

    def __init__(self, filename, sample, transpose=0, speed=1):
        playing = collections.defaultdict(list)
        notes = collections.defaultdict(list)
        # default to acoustic piano
        self.channel_instruments = [sample.instruments[1][0], ] * 16
        self.notecount = 0
        self.maxvolume = 0
        self.maxpitch = 0
        volume = 0

        try:
            with (mido.MidiFile(filename, "r") if isinstance(filename, str) else filename) as mid:
                if mid.type == 2:
                    raise complain.ComplainToUser(
                        "Type 2 (asynchronous) MIDI files are not supported.")

                # label messages from each track
                for track_idx, track in enumerate(mid.tracks):
                    for message in track:
                        # mido hooks __setattr__ so we can't set stuff directly
                        vars(message)["track_idx"] = track_idx

                time = 0
                for message in mid:
                    time += message.time / speed
                    time_samples = int(round(time * sample.framerate))

                    if message.type == "note_on":  # ugh, string-typing
                        if message.channel == 10:
                            try:
                                instrument = sample.percussion[message.note]
                                playing[message.note].append(
                                    Note(start=time_samples,
                                         volume=message.velocity * instrument.volume,
                                         instrument=instrument,
                                         percussion=True))
                            except KeyError:
                                print(
                                    "Warning: Percussion note number outside typical 35-81 range: {}".format(message.note))
                        else:
                            instrument = self.channel_instruments[
                                message.channel]
                            playing[message.note].append(
                                Note(start=time_samples,
                                     volume=message.velocity * instrument.volume,
                                     pitch=message.note + transpose,
                                     instrument=instrument))
                        volume += message.velocity * instrument.volume
                        self.maxvolume = max(volume, self.maxvolume)
                        self.maxpitch = max(message.note + transpose,
                                            self.maxpitch)
                    elif message.type == "note_off":
                        try:
                            note = playing[message.note].pop()
                        except IndexError:  # the pop will fail if there aren't any matching notes playing
                            print(
                                "Warning: Note end event with no matching begin event @ {}".format(time))
                        if len(playing[message.note]) == 0:
                            del playing[message.note]
                        note.finalize(time_samples)
                        notes[note.start].append(note)
                        self.notecount += 1
                        volume -= note.volume
                    elif message.type == "program_change":
                        self.channel_instruments[message.channel] = \
                            sample.instruments[message.program + 1][0]
                if len(playing) != 0:
                    print("Warning: The MIDI ended with notes still playing.")
                    for notelist in playing.values():
                        for note in notelist:
                            note.length = int(
                                time * sample.framerate) - note.start
                            self.notecount += 1
                self.notes = sorted(notes.items(), key=operator.itemgetter(0))
                self.length = max(
                    max(note.start + note.length for note in nlist) for _, nlist in self.notes)
                self.maxpitch = note_to_freq(self.maxpitch)
        except IOError:
            raise complain.ComplainToUser(
                "Error opening MIDI file '{}'.".format(filename))
        except IndexError:
            raise complain.ComplainToUser(
                "This MIDI file is broken. Try opening it in MidiEditor (https://meme.institute/midieditor) and saving it back out again.")



# See https://en.wikipedia.org/wiki/General_MIDI#Parameter_interpretations
# Generated by get_instruments.py

instruments = {
    (1, "Acoustic Grand Piano", "piano"),
    (2, "Bright Acoustic Piano", "piano"),
    (3, "Electric Grand Piano", "piano"),
    (4, "Honky-tonk Piano", "piano"),
    (5, "Electric Piano 1", "piano"),
    (6, "Electric Piano 2", "piano"),
    (7, "Harpsichord", "piano"),
    (8, "Clavinet", "piano"),
    (9, "Celesta", "chromatic percussion"),
    (10, "Glockenspiel", "chromatic percussion"),
    (11, "Music Box", "chromatic percussion"),
    (12, "Vibraphone", "chromatic percussion"),
    (13, "Marimba", "chromatic percussion"),
    (14, "Xylophone", "chromatic percussion"),
    (15, "Tubular Bells", "chromatic percussion"),
    (16, "Dulcimer", "chromatic percussion"),
    (17, "Drawbar Organ", "organ"),
    (18, "Percussive Organ", "organ"),
    (19, "Rock Organ", "organ"),
    (20, "Church Organ", "organ"),
    (21, "Reed Organ", "organ"),
    (22, "Accordion", "organ"),
    (23, "Harmonica", "organ"),
    (24, "Tango Accordion", "organ"),
    (25, "Acoustic Guitar (nylon)", "guitar"),
    (26, "Acoustic Guitar (steel)", "guitar"),
    (27, "Electric Guitar (jazz)", "guitar"),
    (28, "Electric Guitar (clean)", "guitar"),
    (29, "Electric Guitar (muted)", "guitar"),
    (30, "Overdriven Guitar", "guitar"),
    (31, "Distortion Guitar", "guitar"),
    (32, "Guitar Harmonics", "guitar"),
    (33, "Acoustic Bass", "bass"),
    (34, "Electric Bass (finger)", "bass"),
    (35, "Electric Bass (pick)", "bass"),
    (36, "Fretless Bass", "bass"),
    (37, "Slap Bass 1", "bass"),
    (38, "Slap Bass 2", "bass"),
    (39, "Synth Bass 1", "bass"),
    (40, "Synth Bass 2", "bass"),
    (41, "Violin", "strings"),
    (42, "Viola", "strings"),
    (43, "Cello", "strings"),
    (44, "Contrabass", "strings"),
    (45, "Tremolo Strings", "strings"),
    (46, "Pizzicato Strings", "strings"),
    (47, "Orchestral Harp", "strings"),
    (48, "Timpani", "strings"),
    (49, "String Ensemble 1", "ensemble"),
    (50, "String Ensemble 2", "ensemble"),
    (51, "Synth Strings 1", "ensemble"),
    (52, "Synth Strings 2", "ensemble"),
    (53, "Choir Aahs", "ensemble"),
    (54, "Voice Oohs", "ensemble"),
    (55, "Synth Choir", "ensemble"),
    (56, "Orchestra Hit", "ensemble"),
    (57, "Trumpet", "brass"),
    (58, "Trombone", "brass"),
    (59, "Tuba", "brass"),
    (60, "Muted Trumpet", "brass"),
    (61, "French Horn", "brass"),
    (62, "Brass Section", "brass"),
    (63, "Synth Brass 1", "brass"),
    (64, "Synth Brass 2", "brass"),
    (65, "Soprano Sax", "reed"),
    (66, "Alto Sax", "reed"),
    (67, "Tenor Sax", "reed"),
    (68, "Baritone Sax", "reed"),
    (69, "Oboe", "reed"),
    (70, "English Horn", "reed"),
    (71, "Bassoon", "reed"),
    (72, "Clarinet", "reed"),
    (73, "Piccolo", "pipe"),
    (74, "Flute", "pipe"),
    (75, "Recorder", "pipe"),
    (76, "Pan Flute", "pipe"),
    (77, "Blown bottle", "pipe"),
    (78, "Shakuhachi", "pipe"),
    (79, "Whistle", "pipe"),
    (80, "Ocarina", "pipe"),
    (81, "Lead 1 (square)", "synth lead"),
    (82, "Lead 2 (sawtooth)", "synth lead"),
    (83, "Lead 3 (calliope)", "synth lead"),
    (84, "Lead 4 (chiff)", "synth lead"),
    (85, "Lead 5 (charang)", "synth lead"),
    (86, "Lead 6 (voice)", "synth lead"),
    (87, "Lead 7 (fifths)", "synth lead"),
    (88, "Lead 8 (bass + lead)", "synth lead"),
    (89, "Pad 1 (new age)", "synth pad"),
    (90, "Pad 2 (warm)", "synth pad"),
    (91, "Pad 3 (polysynth)", "synth pad"),
    (92, "Pad 4 (choir)", "synth pad"),
    (93, "Pad 5 (bowed)", "synth pad"),
    (94, "Pad 6 (metallic)", "synth pad"),
    (95, "Pad 7 (halo)", "synth pad"),
    (96, "Pad 8 (sweep)", "synth pad"),
    (97, "FX 1 (rain)", "synth effects"),
    (98, "FX 2 (soundtrack)", "synth effects"),
    (99, "FX 3 (crystal)", "synth effects"),
    (100, "FX 4 (atmosphere)", "synth effects"),
    (101, "FX 5 (brightness)", "synth effects"),
    (102, "FX 6 (goblins)", "synth effects"),
    (103, "FX 7 (echoes)", "synth effects"),
    (104, "FX 8 (sci-fi)", "synth effects"),
    (105, "Sitar", "ethnic"),
    (106, "Banjo", "ethnic"),
    (107, "Shamisen", "ethnic"),
    (108, "Koto", "ethnic"),
    (109, "Kalimba", "ethnic"),
    (110, "Bagpipe", "ethnic"),
    (111, "Fiddle", "ethnic"),
    (112, "Shanai", "ethnic"),
    (113, "Tinkle Bell", "percussive"),
    (114, "Agogo", "percussive"),
    (115, "Steel Drums", "percussive"),
    (116, "Woodblock", "percussive"),
    (117, "Taiko Drum", "percussive"),
    (118, "Melodic Tom", "percussive"),
    (119, "Synth Drum", "percussive"),
    (120, "Reverse Cymbal", "percussive"),
    (121, "Guitar Fret Noise", "sound effects"),
    (122, "Breath Noise", "sound effects"),
    (123, "Seashore", "sound effects"),
    (124, "Bird Tweet", "sound effects"),
    (125, "Telephone Ring", "sound effects"),
    (126, "Helicopter", "sound effects"),
    (127, "Applause", "sound effects"),
    (128, "Gunshot", "sound effects")
}

percussion = {
    (35, "Bass Drum 2", "drumset"),
    (36, "Bass Drum 1", "drumset"),
    (37, "Side Stick/Rimshot", "drumset"),
    (38, "Snare Drum 1", "drumset"),
    (39, "Hand Clap"),
    (40, "Snare Drum 2", "drumset"),
    (41, "Low Tom 2", "drumset"),
    (42, "Closed Hi-hat", "drumset"),
    (43, "Low Tom 1", "drumset"),
    (44, "Pedal Hi-hat", "drumset"),
    (45, "Mid Tom 2", "drumset"),
    (46, "Open Hi-hat", "drumset"),
    (47, "Mid Tom 1", "drumset"),
    (48, "High Tom 2", "drumset"),
    (49, "Crash Cymbal 1", "drumset"),
    (50, "High Tom 1", "drumset"),
    (51, "Ride Cymbal 1", "drumset"),
    (52, "Chinese Cymbal"),
    (53, "Ride Bell"),
    (54, "Tambourine"),
    (55, "Splash Cymbal", "drumset"),
    (56, "Cowbell"),
    (57, "Crash Cymbal 2", "drumset"),
    (58, "Vibra Slap"),
    (59, "Ride Cymbal 2", "drumset"),
    (60, "High Bongo"),
    (61, "Low Bongo"),
    (62, "Mute High Conga"),
    (63, "Open High Conga"),
    (64, "Low Conga"),
    (65, "High Timbale"),
    (66, "Low Timbale"),
    (67, "High Agogô", "High Agogo"),
    (68, "Low Agogô", "Low Agogo"),
    (69, "Cabasa"),
    (70, "Maracas"),
    (71, "Short Whistle"),
    (72, "Long Whistle"),
    (73, "Short Güiro", "Short Guiro"),
    (74, "Long Güiro", "Long Guiro"),
    (75, "Claves"),
    (76, "High Wood Block"),
    (77, "Low Wood Block"),
    (78, "Mute Cuíca", "Mute Cuica"),
    (79, "Open Cuíca", "Open Cuica"),
    (80, "Mute Triangle"),
    (81, "Open Triangle")
}
